/*
 * dvdplayer.c: DVD player for DVBs PCI card
 *
 * Copyright (C) 2000 Matjaz Thaler
 *
 * CONTRIBUTORS:
 *  David Toth <dave@IRIS.geobio.elte.hu> 18-7-2000
 *        Communication between the player and the subtitle child 
 *        process.
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at matjaz.thaler@rd.iskraemeco.si 
 *
 * the project's page is at http://www2.arnes.si/~mthale1/index.html
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>		// semaphore 
#include <sys/ipc.h>		// semaphore, shared_memory
#include <sys/sem.h>		// semaphore
#include <sys/shm.h>		// shared memory
#include <sys/wait.h>		// waitpid
#include <signal.h>		// sigaction
#include <string.h>

#include <videodev.h>
#include "css-descramble.h"
#include "css-init.h"
#include "subtitles.h"
#include "remote.h"
#include "dvdplayer.h"

#include "ac3dec/ac3.h"

void usage();
void synchronize(int file, unsigned char *buffer);
int setup_sem_shm_SIGINT();
void data_to_child(unsigned char*, int);
void interrupt_handler();
void change_file(char *fileName, int *file);
int next_file(char *file_name, int *file, int next_file);
int file_seek(int dev, char *file_name, int *file, int *go_to);
void exit_(int status);

unsigned char title_key[5];
unsigned char sector[2048];
int inputFile=0;        // stdin
int videoDev=-1;
int semId=-1;		/* semaphore ID for subtitle child process */
struct data_buf *data_buffer;
int subtitlepid=-1;
int remotepid=-1;
char devVideo[128];
FILE *ac3pipe;

ac3_config_t ac3_config;

int main(int argc, char **argv){
  int c, i;
  struct video_play_mode pmode;

  if(setup_sem_shm_SIGINT()<0)
    exit_(-1);
  data_buffer->video_mode=VIDEO_MODE_PAL;
  data_buffer->css=-1;
  data_buffer->ac3=-1;
  data_buffer->ac3stat=AC3_PLAY;
  data_buffer->lpcm=-1;
  data_buffer->skip=0;
  data_buffer->subtitles=-1;
  data_buffer->subtitleStat=SUBTITLE_DISPLAY;
  data_buffer->ac3sync=-1;
  data_buffer->remote=0;
  data_buffer->changeFile=CURRENT_FILE;
  data_buffer->port=-1;
  strcpy(data_buffer->directory, "/video");
  strcpy(devVideo, "/dev/video");
  strcpy(data_buffer->ac3player, "");

  while(1){
    c = getopt(argc, argv, "va:l:cnf:g:s:j:rd:p:o:h::");
    if (c == -1)
      break;
    switch (c){
    case 'v':
      data_buffer->video_mode=-1;
      fprintf (stderr, "do not playback video \n");
      break;
    case 'a':
      data_buffer->lpcm=-1; 
      if(optarg[0]=='a'){
	data_buffer->ac3=-2;
	fprintf (stderr, "ac3 sound\n");
      }
      else{
	sscanf(optarg, "%d", &data_buffer->ac3);
	data_buffer->ac3 &= 7;
	fprintf (stderr, "ac3 soundtrack %d\n", data_buffer->ac3);
      }
      break;
    case 'l':
      data_buffer->ac3=-1;
      if(optarg[0]=='a'){
	data_buffer->lpcm=-2;
	fprintf (stderr, "lpcm sound\n");
      }
      else{
	sscanf(optarg, "%d", &data_buffer->lpcm);
	data_buffer->lpcm &= 7;
	fprintf (stderr, "lpcm soundtrack %d\n", data_buffer->lpcm);
      }
      break;
    case 'c':
      data_buffer->css=1;
      fprintf (stderr, "css decode\n");
      break;
    case 'n':
      data_buffer->video_mode=VIDEO_MODE_NTSC;
      fprintf (stderr, "NTSC playback\n");
      break;
    case 'f':
      strcpy(data_buffer->fileName, optarg);
      inputFile=-1;
      break;
    case 'g':
      sscanf(optarg, "%d", &data_buffer->skip);
      fprintf (stderr, "skip %d MB\n", data_buffer->skip);
      break;
    case 's':
      sscanf(optarg, "%d", &data_buffer->subtitles);
      data_buffer->subtitles &= 31;
      fprintf (stderr, "subtitles %d\n", data_buffer->subtitles);
      break;
    case 'j':
      sscanf(optarg, "%d", &data_buffer->ac3sync);
      if(data_buffer->ac3sync>16)
	data_buffer->ac3sync=16;
      fprintf (stderr, "skip %d/1024 sound samples\n", data_buffer->ac3sync);
      break;
    case 'r':
      data_buffer->remote=1;
      fprintf (stderr, "use LIRC\n");
      if(data_buffer->port >= 0){
	fprintf (stderr, "you can use only one option: LIRC or network connection\n");
	exit_(1);
      }
      break;
    case 'p':
      sscanf(optarg, "%d", &data_buffer->port);
      fprintf (stderr, "listening on port %d\n", data_buffer->port);
      if(data_buffer->remote == 1){
	fprintf (stderr, "you can use only one option: LIRC or network connection\n");
	exit_(1);
      }
      break;
    case 'd':
      strcpy(devVideo, optarg);
      break;
    case 'o':
      strcpy(data_buffer->ac3player, optarg);
      fprintf (stderr, "ac3 decoder: %s\n", data_buffer->ac3player);
      break;
    case 'h':
    default:
      usage();
      exit_(1);
    }
  }
  // init
  if(data_buffer->css==1){
    if(css_init(title_key))
      exit_(1);
  }
  if((data_buffer->video_mode >= 0) || (data_buffer->subtitles >= 0) || data_buffer->remote==1 || data_buffer->port > 0){
    videoDev=open(devVideo, O_RDWR);
    if(videoDev < 0){
      fprintf (stderr, "can't open %s\n", devVideo);
      exit_(1);
    }
    else
      fprintf (stderr, "opening %s\n", devVideo);
  }
  if(data_buffer->remote > 0 || data_buffer->port > 0){
    if(!(remotepid=fork()))
    {
      remote_control(videoDev);
      exit_(0);
    }
    if(remotepid<0)
    {
      fprintf(stderr,"ERROR: fork (remote child process) failed!\n");
      close(videoDev);
      exit_(-1);
    }
  }
  if(data_buffer->ac3 != -1 || data_buffer->remote==1 || data_buffer->port>0){
    
    ac3_config.num_output_ch = 2;
    ac3_config.flags = 0;
    ac3_init(&ac3_config);
  }

  if(data_buffer->subtitles >= 0 || data_buffer->remote==1 || data_buffer->port>0)
  {
    if(!(subtitlepid=fork()))
    {
      subtitles_decode(videoDev);
      exit_(0);
    }
    if(subtitlepid<0)
    {
      fprintf(stderr,"ERROR: fork (subtitle child process) failed!\n");
      close(videoDev);
      exit_(-1);
    }
  }
  if(data_buffer->ac3player != NULL){
    ac3pipe=popen(data_buffer->ac3player, "w");
    if(ac3pipe==NULL){
      fprintf (stderr, "can't run %s\n", data_buffer->ac3player);
      exit_(-1);
    }
  }
  if(inputFile!=0){
    inputFile=open(data_buffer->fileName, O_RDONLY);
    if(inputFile<0){
      fprintf (stderr, "can't open %s\n", data_buffer->fileName);
      exit_(1);
    }
    fprintf (stderr, "input file: %s\n", data_buffer->fileName);
  }
  if(data_buffer->skip && inputFile)
    file_seek(videoDev, data_buffer->fileName, &inputFile, &data_buffer->skip);

  if(data_buffer->video_mode == VIDEO_MODE_NTSC){
    pmode.mode = VID_PLAY_VID_OUT_MODE;
    pmode.p1=data_buffer->video_mode;
    ioctl(videoDev, VIDIOCSPLAYMODE, &pmode);
  }
  
  while(1){                                                    //playback
    if(data_buffer->changeFile != CURRENT_FILE){
      if(data_buffer->changeFile == NEW_FILE)
	change_file(data_buffer->fileName, &inputFile);
      else{
	if(!next_file(data_buffer->fileName, &inputFile, data_buffer->changeFile)){
	  data_buffer->changeFile=CURRENT_FILE;
	  continue;
	}
      }
      data_buffer->changeFile=CURRENT_FILE;
      pmode.mode = VID_PLAY_RESET;
      ioctl(videoDev, VIDIOCSPLAYMODE, &pmode);
      if(data_buffer->video_mode == VIDEO_MODE_NTSC){
	pmode.mode = VID_PLAY_VID_OUT_MODE;
	pmode.p1=data_buffer->video_mode;
	ioctl(videoDev, VIDIOCSPLAYMODE, &pmode);
      }
      data_buffer->ac3stat=AC3_START;
    }
    if(data_buffer->go_to != 0){
      file_seek(videoDev, data_buffer->fileName, &inputFile, &data_buffer->go_to);
      if(data_buffer->video_mode == VIDEO_MODE_NTSC){
	pmode.mode = VID_PLAY_VID_OUT_MODE;
	pmode.p1=data_buffer->video_mode;
	ioctl(videoDev, VIDIOCSPLAYMODE, &pmode);
      }
      data_buffer->ac3stat=AC3_START;
    }
    if(read(inputFile, sector, 2048) == 2048){                 //read input file
      if (!(sector[0]==0x00 && sector[1]==0x00 && sector[2]==0x01 && sector[3]==0xba)) {
	synchronize(inputFile, sector);
      }
      if ((sector[14+3]&0xF0)==0xE0 || (sector[14+3]&0xF0)==0xC0){ // video, mpeg audio
	if(data_buffer->video_mode>=0){
	  if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	    css_descramble(sector, title_key);
	    sector[20] &= 0x8f;
	  }
	  write(videoDev, sector+14 , sector[19]+(sector[18]<<8)+6);
	}
      }
      else if (((data_buffer->ac3>=0) && ((sector[14+sector[14+8]+9])==(0x80|data_buffer->ac3))) || 
	       ((data_buffer->ac3==-2) && ((sector[14+sector[14+8]+9]&0xF8)==0x80))) {  // Dolby AC3 audio
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	if(strcmp(data_buffer->ac3player, "") && (data_buffer->ac3stat != AC3_STOP))
	  fwrite(sector+sector[22]+27, ((sector[18]<<8)|sector[19])-sector[22]-7, 1, ac3pipe);
	else if(data_buffer->ac3stat == AC3_PLAY)
	  ac3_decode_data(sector+sector[22]+27, sector+20+((sector[18]<<8)|sector[19]), videoDev, 0, data_buffer->ac3sync);
	else if(data_buffer->ac3stat == AC3_START){
	  ac3_decode_data(sector+sector[22]+27, sector+20+((sector[18]<<8)|sector[19]), videoDev, 1, data_buffer->ac3sync);
	  data_buffer->ac3stat = AC3_PLAY;
	}
      }
      else if (((data_buffer->lpcm>=0) && ((sector[14+sector[14+8]+9])==(0xA0|data_buffer->lpcm))) || 
	       ((data_buffer->lpcm==-2) && ((sector[14+sector[14+8]+9]&0xF8)==0xA0))) {  // LPCM audio
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	write(videoDev, sector+14 , sector[19]+(sector[18]<<8)+6);
      }
      else if ((data_buffer->subtitles>=0) && (sector[14+sector[14+8]+9])==
	       (0x20|data_buffer->subtitles)) {                           //spu data
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	i = 24 + sector[22];
	data_to_child(sector+i , 2048-i);
      }
    }
    else{                                                   // end of file
      if(inputFile!=0){
	if(next_file(data_buffer->fileName, &inputFile, 1))
	  continue;
	close(inputFile);
      }
      if(data_buffer->remote || data_buffer->port>0){                              // select new file
	pmode.mode = VID_PLAY_RESET;
	ioctl(videoDev, VIDIOCSPLAYMODE, &pmode);
	while(1){
	  if(data_buffer->exit_child)
	    break;
	  sleep(1);
	  if(data_buffer->changeFile != CURRENT_FILE){
	    inputFile=open(data_buffer->fileName, O_RDONLY);              
	    if(inputFile<0){
	      fprintf (stderr, "can't open %s\n", data_buffer->fileName);
	      data_buffer->changeFile=CURRENT_FILE;
	    }
	    else{
	      fprintf (stderr, "playing: %s\n", data_buffer->fileName);
	      break;
	    }
	  }
	}
      }
      else
	break;
    }
    if(data_buffer->exit_child) 
      break;
  }
  exit_(0);
  return 0;
} 


void usage(){
  fprintf (stderr, "\n");
  fprintf (stderr, "Usage:\tdvdplayer [OPTION]...\n");
  fprintf (stderr, "OPTIONS:\n");
  fprintf (stderr, " -f filename\topen file [stdin]\n");
  fprintf (stderr, " -c\t\tcss decode (playback from DVD)\n");
  fprintf (stderr, " -n\t\tNTSC playback\n");
  fprintf (stderr, " -a channel\tplay ac3 sound\n");
  fprintf (stderr, " -l channel\tplay lpcm sound\n");
  fprintf (stderr, " -s subtitle  \tdisplay subtitles\n");
  fprintf (stderr, " -v\t\tdo not play video\n");
  fprintf (stderr, " -g n\t\tstart playback at n M bytes of input file\n");
  fprintf (stderr, " -j n\t\tskip n/1024 sound samples - this option\n");
  fprintf (stderr, "    \t\tshould remove video disturbances\n");
  fprintf (stderr, " -r \t\tremote control with LIRC\n");
  fprintf (stderr, " -p port\tcontrol via a network connection\n");
  fprintf (stderr, " -d dev\t\tvideo device    [/dev/video]\n");
  fprintf (stderr, " -o decoder\tsend ac3 stream to decoder (ac3play, ac3spdif)\n");
  fprintf (stderr, " -h\t\tshow this usage information\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "EXAMPLE:\n");
  fprintf (stderr, " -playback from DVD:\n");
  fprintf (stderr, "   dvdplayer -a 0 -c -s 0 -f /dvd/video_ts/vts_01_1.vob\n");
  fprintf (stderr, " -playback from hard disk:\n");
  fprintf (stderr, "   dvdplayer -a 0 -s 0 -f foo.vob\n");
  fprintf (stderr, "\n");
}

void synchronize(int file, unsigned char *buffer){

  int sync=0;
  
  fprintf(stderr, "sync\n");
  while(sync==0){
    while(buffer[0] != 0)
      read(file, &buffer[0], 1);
    read(file, &buffer[1], 1);
    if(buffer[1] != 0)
      continue;
    buffer[2]=0;
    while(buffer[2] == 0)
      read(file, &buffer[2], 1);
    if(buffer[2] != 1)
      continue;
    read(file, &buffer[3], 1);
    if(buffer[3] != 0xba)
      continue;
    read(file, &buffer[4], 2044);
    sync=1;
  }
}

int setup_sem_shm_SIGINT()
{
  union semun arg;
  int shmid_buf;
  struct sigaction new_sa, old_sa;

  // setup a new interrupt handler
  new_sa.sa_handler = interrupt_handler;
  sigemptyset(&new_sa.sa_mask);
  new_sa.sa_flags = 0;
  if (sigaction(SIGINT, &new_sa, &old_sa) == -1) 
    return (-1);

  // create a semaphore
  if((semId = semget(IPC_PRIVATE, 2, 0600)) < 0){
    fprintf(stderr,"ERROR: semget failed, errno is"); perror(" ");
    return(-1);
  }

  arg.val=0;
  if(semctl(semId, 0, SETVAL, arg) == -1){
    semctl(semId, 1, IPC_RMID, arg);	// remove semaphore
    semId=-1;
    fprintf(stderr,"ERROR: semctl failed, errno is"); perror(" ");
    return(-1);
  }

  // create shared memory 
  if((shmid_buf = shmget(IPC_PRIVATE, sizeof(struct data_buf),
                         IPC_CREAT|0600)) == -1){
    semctl(semId, 1, IPC_RMID, arg);	/* remove semaphore */
    semId=-1;
    fprintf(stderr,"ERROR: shmget failed, errno is"); perror(" ");
  }
  data_buffer = (struct data_buf*) shmat(shmid_buf, 0, 0);
  shmctl(shmid_buf, IPC_RMID, 0);
  
  // init buffers
  data_buffer->exit_child=0;
  data_buffer->input_pointer=0;
  data_buffer->output_pointer=0;
  data_buffer->input_pointer_ac3=0;
  data_buffer->output_pointer_ac3=0;
  return(0);
}

void data_to_child(unsigned char *data, int datasize)
{
  while(datasize){
    if( ((data_buffer->input_pointer+1)%SUBTITLE_BUFFER_SIZE) !=
	data_buffer->output_pointer){    // There is room in the buffer
      data_buffer->data[data_buffer->input_pointer]=*data++;
      datasize--;
      data_buffer->input_pointer=
	(data_buffer->input_pointer+1)%SUBTITLE_BUFFER_SIZE;
    }
    else{                           // Buffer is full
      if(sem_ctl(semId, 0, 0))
	return;
      if(sem_op(semId, 0, -1))      // sleep until child is done
	return;
    }
  }
  if(sem_ctl(semId, 0, 0))          // wake up child if sleeps
    return;
}

void interrupt_handler()
{
  if(semId>=0) semctl(semId, 1, IPC_RMID, NULL);	/* remove semaphore */
  semId=-1;
  if(!data_buffer->exit_child) 
    fprintf(stderr,"Interrupt signal received.\n");
  data_buffer->exit_child=1;
}

void change_file(char *fileName, int *file){

  int f;

  f=open(fileName, O_RDONLY);
  if(f<0)
    fprintf (stderr, "can't open %s\n", fileName);
  else{
    close(*file);
    *file=f;
  }
}

int next_file(char *file_name, int *file, int next_file){
  
  int len=strlen(file_name);
  int f;
  
  if(next_file){
    if(!(file_name[len-5]>='1' && file_name[len-5]<='9' && len>=12))
      return 0;
    file_name[len-5]+=next_file;
    if(!(file_name[len-5]>='1' && file_name[len-5]<='9' && len>=12)){
      file_name[len-5]-=next_file;
      return 0;
    }
    f=open(file_name, O_RDONLY);
    if(f<0){
      fprintf (stderr, "can't open %s\n", file_name);
      file_name[len-5]-=next_file;
      return 0;
    }
    else{
      fprintf (stderr, "opening %s\n", file_name);
      close(*file);
      *file=f;
      return 1;
    }
  }
  return 0;
}

int file_seek(int dev, char *file_name, int *file, int *go_to){

  struct video_play_mode pmode;
  long long int i, end, cur, seek;

  seek=*go_to;
  seek*=1024*1024;
  cur=lseek(*file, 0, SEEK_CUR);
  end=lseek(*file, 0, SEEK_END);
  while(seek){
    i=cur+seek;
    if(!(i<0 || i>end)){
      lseek(*file, i, SEEK_SET);
      seek=0;
    }
    else if(i<0){
      if(next_file(file_name, file, -1)){
	seek+=cur;
	cur=end=lseek(*file, 0, SEEK_END);
      }
      else{
	lseek(*file, 0, SEEK_SET);
	break;
      }
    }
    else{
      if(next_file(file_name, file, 1)){
	seek-=end-cur;
	cur=0;
	end=lseek(*file, 0, SEEK_END);
      }
      else{
	lseek(*file, -4*1024*1024, SEEK_END);
	break;
      }
    }
  }
  *go_to=0;
  pmode.mode = VID_PLAY_RESET;
  ioctl(dev, VIDIOCSPLAYMODE, &pmode);
  if(seek==0)
    return 1;
  else
    return 0;
}

int sem_ctl(int semid, int semnum, int val){

  union semun arg;

  arg.val=val;	
  if(semctl(semid, semnum, SETVAL, arg) == -1){
    if(!data_buffer->exit_child){
      fprintf(stderr,"ERROR: semctl failed, errno is"); perror(" ");
      data_buffer->exit_child=1;
    }
    return 1;
  }
  return 0;
}

int sem_op(int semid, int semnum, int op){

  struct sembuf semBuf;

  semBuf.sem_num = semnum;
  semBuf.sem_op  = op; 
  semBuf.sem_flg = 0;
  if(semop(semid,&semBuf,1)){
    if(!data_buffer->exit_child){
      fprintf(stderr,"ERROR: semop failed, errno is"); perror(" ");
      data_buffer->exit_child=1;
    }
    return 1;
  }
  return 0;
}

void exit_(int status){

  data_buffer->exit_child=1;
  usleep(100000);
  if(subtitlepid > 0)
    kill(subtitlepid, 15);
  if(remotepid > 0)
    kill(remotepid, 15);
  if(videoDev >= 0)
    close(videoDev);
  if(inputFile > 0)
    close(inputFile);
  if(strcmp(data_buffer->ac3player, ""))
    pclose(ac3pipe);
  if(semId >= 0) 
    semctl(semId, 1, IPC_RMID, NULL);	  //remove semaphore
  exit(status);
}
