/*
 * remote.c: remote control for DVD player
 *
 * Copyright (C) 2000 Matjaz Thaler
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at matjaz.thaler@rd.iskraemeco.si 
 *
 * the project's page is at http://www2.arnes.si/~mthale1/index.html
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <OSD.h>
#include <videodev.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h> 
#include <arpa/inet.h>
#include <netinet/in.h>

#include "dvdplayer.h"
#include "remote.h"

int get_ir();
int getFileName(int dev, char *fileName, char *dir);
void display_menu(int dev, struct data_buf *data_buffer);
void go_to(int dev, struct data_buf *data_buffer, int direction);
int lirc_nextcode(char **code);

struct video_play_mode pmode;
struct sockaddr_un addr;

int sockfd=-1, new_fd=-1;
struct sockaddr_in my_addr;
struct sockaddr_in their_addr;
int sin_size;
int i;
int on=1;
char buff[128];

void remote_control(int dev){
  if(data_buffer->port == -1){
    addr.sun_family=AF_UNIX;
    strcpy(addr.sun_path, "/dev/lircd");
    sockfd=socket(AF_UNIX, SOCK_STREAM, 0);
    if(sockfd == -1){
      fprintf(stderr, "lirc: could not open socket\n");
      data_buffer->exit_child=1;
      exit(EXIT_FAILURE);  
    }
    if(connect(sockfd, (struct sockaddr *)&addr, sizeof(addr))==-1){
      close(sockfd);
      fprintf(stderr, "lirc: could not connect to socket\n");
      data_buffer->exit_child=1;
      exit(EXIT_FAILURE);  
    }
  }
  else{
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      perror("socket");
      exit(1);
    }
    
    my_addr.sin_family = AF_INET;         // host byte order
    my_addr.sin_port = htons(data_buffer->port);     // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // auto-fill with my IP
    bzero(&(my_addr.sin_zero), 8);        // zero the rest of the struct
    
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
    if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
      perror("bind");
      exit(1);
    }
    if (listen(sockfd, 5) == -1) {
      perror("listen");
      exit(1);
    }
  }

  while(get_ir()){
    if(data_buffer->exit_child)
      exit(-1);
    if(!strcmp(data_buffer->ir.command, PLAY)){
      data_buffer->ac3stat=AC3_START;
      if(data_buffer->subtitleStat==SUBTITLE_HIDE)
	data_buffer->subtitleStat=SUBTITLE_DISPLAY;
      if((data_buffer->ac3 != -1  && !strcmp(data_buffer->ac3player, "")) || (data_buffer->lpcm != -1)){
	data_buffer->go_to = -1;
	pmode.mode = VID_PLAY_SLOW_MOTION;
	ioctl(dev, VIDIOCSPLAYMODE, &pmode);
      }
      else{
      pmode.mode = VID_PLAY_NORMAL;
      ioctl(dev, VIDIOCSPLAYMODE, &pmode);
      }
    }
    else if(!strcmp(data_buffer->ir.command, PAUSE)){
      if(data_buffer->ac3stat==AC3_PLAY)
	data_buffer->ac3stat=AC3_STOP;
      pmode.mode = VID_PLAY_PAUSE;
      ioctl(dev, VIDIOCSPLAYMODE, &pmode); 
    }
    else if(!strcmp(data_buffer->ir.command, SLOW)){
      if(data_buffer->ac3stat==AC3_PLAY)
	data_buffer->ac3stat=AC3_STOP;
      pmode.mode = VID_PLAY_SLOW_MOTION;
      ioctl(dev, VIDIOCSPLAYMODE, &pmode); 
    }
    else if(!strcmp(data_buffer->ir.command, FASTFORWARD)){
      if(data_buffer->ac3stat==AC3_PLAY)
	data_buffer->ac3stat=AC3_STOP;
      if(data_buffer->subtitleStat==SUBTITLE_DISPLAY)
	data_buffer->subtitleStat=SUBTITLE_HIDE;
      pmode.mode = VID_PLAY_FAST_FORWARD;
      ioctl(dev, VIDIOCSPLAYMODE, &pmode); 
    }
    else if(!strcmp(data_buffer->ir.command, MENU))
      display_menu(dev, data_buffer);
    else if(!strcmp(data_buffer->ir.command, LEFT))
      go_to(dev, data_buffer, PREW_FILE);
    else if(!strcmp(data_buffer->ir.command, RIGHT))
      go_to(dev, data_buffer, NEXT_FILE);
  }
  data_buffer->exit_child=1;
  exit(0);
}

int get_ir(){

  char *buf;
  char remote[20];
  unsigned int repeat;


  if(data_buffer->port > 1){
  new_connecton:
    if(new_fd<0){
      sin_size = sizeof(struct sockaddr_in);
      if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size)) == -1) {
	perror("accept");
	return 0;
      }
      printf("server: got connection from %s\n", inet_ntoa(their_addr.sin_addr));
      if (send(new_fd, "Connected to dvdplayer\n", 23, 0) == -1)
	perror("send");
    }
    while(1){
      i=0;
      buff[i]=0;
      while(buff[i]==0 || buff[i]==0xa || buff[i]==0xd){
	read(new_fd, &buff[i], 1);
      }
      while(!(buff[i]==0 || buff[i]==0xa || buff[i]==0xd)){
	i++;
	read(new_fd, &buff[i], 1);
      }
      buff[i]=0;
      //printf("%s\n", buff);
      if(!strcmp(buff, "quit")){
	close(new_fd);
	new_fd=-1;
      goto new_connecton;
      }
      else{
	strcpy(data_buffer->ir.command, buff);
	if(data_buffer->ir.command[1]==0x0 && data_buffer->ir.command[0]>0x2f && data_buffer->ir.command[0]<0x3a){
	  data_buffer->ir.number=data_buffer->ir.command[0]-0x30;
	  data_buffer->ir.state = NUM_KEY;
	}
	else
	  data_buffer->ir.state = COMMAND;
	//printf("command: %s\n", data_buffer->ir.command);
	return 1;
      }
    }
  }
  
  else{
    while(1){
      if(lirc_nextcode(&buf)==-1)
	return 0;
      sscanf(buf, "%*s %x %s %s", &repeat, data_buffer->ir.command, remote);
      //printf("%d %s %s\n", repeat, remote, data_buffer->ir.command);
      if(!strcmp(remote, REMOTE_NAME)){
	if(data_buffer->ir.command[1]==0x0 && data_buffer->ir.command[0]>0x2f && data_buffer->ir.command[0]<0x3a){
	  data_buffer->ir.number=data_buffer->ir.command[0]-0x30;
	  data_buffer->ir.state = NUM_KEY;
	}
	else
	  data_buffer->ir.state = COMMAND;
      }
      if(!(repeat%4))
	return 1;
    }
  }
}

int lirc_nextcode(char **code){

#define PACKET_SIZE 128

  int packet_size=PACKET_SIZE;
  int end_len=0;
  ssize_t len=0;
  char *end,c;
  char lirc_buffer[PACKET_SIZE+1];

  *code=NULL;
  while((end=strchr(lirc_buffer,'\n'))==NULL){
    len=read(sockfd, lirc_buffer+end_len, packet_size-end_len);
    end_len+=len;
    lirc_buffer[end_len]=0;
    if((end=strchr(lirc_buffer, '\n'))==NULL){
      return(0);
    }
  }
  end++;
  end_len=strlen(end);
  c=end[0];
  end[0]=0;
  *code=strdup(lirc_buffer);
  end[0]=c;
  memmove(lirc_buffer,end,end_len+1);
  if(*code==NULL) 
    return(-1);
  return(0);
}

int getFileName(int dev, char *fileName, char *dir){

#define  LINES_NO    15       
#define  FONT_SIZE    1     
#define  HEIGHT      26  


  int i, k;
  char stmp[512];
  char btmp[512];
  char type[128];
  char name[128][64];
  char dirstrc[16][32];
  int dirnum;
  int fnum=0;
  FILE *p; 
  int start, size, cur;
  
  k=0;
  i=1;
  dirnum=0;
  while(dir[i] != 0){
    if(dir[i] == '/'){
      dirstrc[dirnum][k]=0;
      dirnum++;
      k=0;
      i++;
    }
    else{
      dirstrc[dirnum][k]=dir[i];
      i++;
      k++;
    }
  }
  dirstrc[dirnum][k]=0;
  if(i>1)
    dirnum++;
  
  while(1){
    start=0;
    size=LINES_NO;
    cur=0;
    stmp[0]=0;
    if(dirnum == 0)
      sprintf(stmp,"/");
    else for(i=0; i < dirnum; i++)
      sprintf(stmp,"%s/%s", stmp, dirstrc[i]);
    sprintf(btmp,"ls -lt %s | grep r", stmp);
    strcpy(dir, stmp);
    p = popen(btmp, "r");
    i=0;
    if (p) {
      while(fgets(btmp, 512, p) && sscanf(btmp, "%s %*s %*s %*s %*s %*s %*s %*s %s", 
					  stmp, btmp) == 2){
	sscanf(stmp, "%c", &type[i]);
	sscanf(btmp, "%s", &name[i][0]);
	i++;
	if(i >= 128)
	  break;
      }
      fnum = i;
      pclose(p);
    }
    else
      fprintf(stderr, "Error while opening pipe!");
    size=LINES_NO;
    if(fnum<LINES_NO)
      size=fnum;
  Replot:
    OSDClose(dev);
    OSDOpen(dev, 100, 30, 620, 456, 2, 0);
    OSDSetColor(dev, 0, 0, 0, 0, 160);
    OSDSetColor(dev, 1, 0, 255, 255, 255);
    OSDSetColor(dev, 2, 255, 0, 255, 255);
    OSDSetColor(dev, 3, 255, 255, 0, 255);
    //OSDFill(dev, 0);
    OSDText(dev, 50, 7, FONT_SIZE, 1, " ../");
    sprintf(stmp,"directory:  %s", dir);
    OSDText(dev, 120, 7, FONT_SIZE, 2, stmp);
    for(i=0; i<size; i++){
      sprintf(btmp, " %s", name[start+i]);
      if(type[start+i] == 'd')
	OSDText(dev, 50, 7+HEIGHT*(i+1), FONT_SIZE, 1, btmp);
      else
	OSDText(dev, 50, 7+HEIGHT*(i+1), FONT_SIZE, 3, btmp);
    }
    if(fnum == 0)
      cur=-1;
    i=cur;
    if(type[i+start] == 'd' || i==-1)
      OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 1, " >>");
    else
      OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 3, " >>");
    while(get_ir()){
      if(!strcmp(data_buffer->ir.command, ENTER)){
	//fprintf(stderr, "name %s \n", name[i+start]);
	if(i == -1){
	  if(dirnum > 0)
	    dirnum--;
	}
	else if(type[i+start] == 'd'){
	  dirnum++;
	  sprintf(dirstrc[dirnum-1], "%s", name[i+start]);
	}
	else{
	  stmp[0]=0;
	  for(k=0; k < dirnum; k++)
	    sprintf(stmp,"%s/%s", stmp, dirstrc[k]);
	  sprintf(stmp,"%s/%s", stmp, name[i+start]);
	  //printf("%s\n", stmp);
	  OSDClose(dev);
	  sprintf(fileName, "%s", stmp);
	  return 1;
	}
	OSDClose(dev);
	break;
      }
      else if(!strcmp(data_buffer->ir.command, UP) || !strcmp(data_buffer->ir.command, LEFT)){
	OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 0, " >>");
	if(!strcmp(data_buffer->ir.command, LEFT))
	  i-=LINES_NO-1;
	i--;
	if(i<-1){
	  if(start == 0)
	    i=-1;
	  else{
	    start-=LINES_NO-1;
	    cur=LINES_NO-2;
	    if(start<0){
	      cur=start;
	      start=0;
	      cur+=LINES_NO-2;
	    }
	    goto Replot;
	  }
	}
	if(type[i+start] == 'd' || i == -1)
	  OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 1, " >>");
	else
	  OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 3, " >>");
      }
      else if(!strcmp(data_buffer->ir.command, DOWN) || !strcmp(data_buffer->ir.command, RIGHT)){
	OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 0, " >>");
	if(!strcmp(data_buffer->ir.command, RIGHT))
	  i+=LINES_NO-1;
	i++;
	if(i>=size){
	  if(start+size == fnum)
	    i=size-1;
	  else{
	    start+=LINES_NO-1;
	    cur=1;
	    if(start>fnum-size){
	      cur=start;
	      start=fnum-size;
	      cur-=start-1;
	    }
	    goto Replot;
	  }
	}
	if(type[i+start] == 'd' || i == -1)
	  OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 1, " >>");
	else
	  OSDText(dev, 0, 7+HEIGHT*(i+1), FONT_SIZE, 3, " >>");
      }
      else if(!strcmp(data_buffer->ir.command, MENU) || !strcmp(data_buffer->ir.command, NORMAL)){
	OSDClose(dev);
	return 0;
      }
    }
  }
}

void display_line(int dev, int position, int bold, struct data_buf *data_buffer){

#define  FONT_SIZE_    1     
#define  HEIGHT_      28
#define  LEFT_POS     20
#define  RIGHT_POS   300

  int color=7|(4<<16);
  char buff[128];

  if(bold)
    color=6|(4<<16);
  switch (position){
   case 0:
    OSDText(dev, (LEFT_POS+RIGHT_POS)/2, 7, FONT_SIZE_, 5 | (4<<16), "SETUP");
    break;
   case MENU_VIDEO:
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "VIDEO");
    if(data_buffer->video_mode==-1)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "OFF    ");
    else if(data_buffer->video_mode==VIDEO_MODE_PAL)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "PAL   ");
    else if(data_buffer->video_mode==VIDEO_MODE_NTSC)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "NTSC  ");
    break;
  case MENU_SOUND:
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "SOUNDTRACK");
    if(data_buffer->ac3==-1 && data_buffer->lpcm==-1)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "OFF     ");
    else if(data_buffer->ac3 >= 0){
      sprintf(buff, "AC3  %d     ", data_buffer->ac3);
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, buff);
    }
    else if(data_buffer->lpcm >= 0){
      sprintf(buff, "LPCM %d     ", data_buffer->lpcm);
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, buff);
    }
    break;
   case MENU_SUBTITLES:
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "SUBTITLES");
    if(data_buffer->subtitles==-1)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "OFF");
    else{
      sprintf(buff, "%3d     ", data_buffer->subtitles);
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, buff);
    }
    break;
  case MENU_CSS:
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "CSS DECODE");
    if(data_buffer->css == 1)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "ON   ");
    else
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "OFF  ");
    break;
   case MENU_SYNC:
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "SKIP SOUND SAMP.");
    if(data_buffer->ac3sync==-1)
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, "OFF  ");
    else{
      sprintf(buff, "%3d     ", data_buffer->ac3sync);
      OSDText(dev, RIGHT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, buff);
    }
    break;
  case MENU_FILE:
    sprintf(buff, "FILE:  %s                         ", data_buffer->fileName);
    OSDText(dev, LEFT_POS, 7+HEIGHT_*position, FONT_SIZE_, color, buff);
    break;
  case MENU_EXIT:
    OSDText(dev, (LEFT_POS+RIGHT_POS)/2, 7+HEIGHT_*position, FONT_SIZE_, color, "EXIT");
    break;
  }
}

void display_menu(int dev, struct data_buf *data_buffer){

  int i;
  int pos=1;

  data_buffer->subtitleStat=SUBTITLE_OVERLAY;
  OSDClose(dev);
  OSDOpen(dev, 100, 50, 620, 450, 4, 0);
  OSDSetColor(dev, 0, 0, 0, 0, 0);
  OSDSetColor(dev, 4, 0, 0, 0, 160);
  OSDSetColor(dev, 5, 0, 255, 255, 255);
  OSDSetColor(dev, 6, 255, 0, 255, 255);
  OSDSetColor(dev, 7, 255, 255, 0, 255);
  OSDFill(dev, 4); 
  OSDFillBlock(dev, 0, 240, 520, 400, 0);
  for(i=0; i<=MENU_LAST; i++){
    if(i==1)
      display_line(dev, i, 1, data_buffer);
    else
      display_line(dev, i, 0, data_buffer);
  }
  while(get_ir()){
    if(!strcmp(data_buffer->ir.command, UP)){
      display_line(dev, pos, 0, data_buffer);
      if(pos==1)
	pos=MENU_LAST;
      else
	pos--;
      display_line(dev, pos, 1, data_buffer);
    }
    if(!strcmp(data_buffer->ir.command, DOWN)){
      display_line(dev, pos, 0, data_buffer);
      if(pos==MENU_LAST)
	pos=1;
      else
	pos++;
      display_line(dev, pos, 1, data_buffer);
    }
    else if(!strcmp(data_buffer->ir.command, ENTER)){
      if(pos==MENU_FILE){
	data_buffer->subtitleStat=SUBTITLE_HIDE;
	if(getFileName(dev, data_buffer->fileName, data_buffer->directory)){
	  //printf("filename: %s\n", data_buffer->fileName);
	  data_buffer->changeFile=NEW_FILE;
	}
	data_buffer->subtitleStat=SUBTITLE_OVERLAY;
      }
      else if(pos==MENU_EXIT){
	data_buffer->exit_child=1;
	exit(0);
      }
      OSDClose(dev);
      break;
    }
    else if(!strcmp(data_buffer->ir.command, NORMAL) || !strcmp(data_buffer->ir.command, MENU)){
      OSDClose(dev);
      break;
    }
    else if(!strcmp(data_buffer->ir.command, LEFT) || !strcmp(data_buffer->ir.command, RIGHT)){
      if(!strcmp(data_buffer->ir.command, RIGHT))
	i=1;
      else
	i=-1;
      switch (pos){
      case MENU_VIDEO:
	i+=data_buffer->video_mode;
	if(i<-1)
	  i=-1;
	else if(i>1)
	  i=1;
	data_buffer->video_mode=i;
	if(i>=0){
	  pmode.mode = VID_PLAY_VID_OUT_MODE;
	  pmode.p1=data_buffer->video_mode;
	  ioctl(dev, VIDIOCSPLAYMODE, &pmode);
	}
	break;
      case MENU_SOUND:
	if(data_buffer->ac3>=0)
	  i+=data_buffer->ac3+1;
	if(data_buffer->lpcm>=0)
	  i+=data_buffer->lpcm+9;
	if(i>16)
	  i=16;
	if(i<0)
	  i=0;
	i--;
	if(i<=7){
	  data_buffer->ac3=i;
	  data_buffer->ac3stat=AC3_START;
	  data_buffer->lpcm=-1;
	}
	else{
	  data_buffer->lpcm=i-8;
	  data_buffer->ac3=-1;
	}
	break;
      case MENU_SUBTITLES:
	data_buffer->subtitles+=i;
	if(data_buffer->subtitles>31)
	  data_buffer->subtitles=31;
	if(data_buffer->subtitles<-1)
	  data_buffer->subtitles=-1;
	break;
      case MENU_CSS:
	if(data_buffer->css != -1)
	  data_buffer->css=(~data_buffer->css)&1;
	else
	  fprintf(stderr, "css decode: run dvdplayer with the command line option -c\n");
	break;
      case MENU_SYNC:
	data_buffer->ac3sync+=i;
	if(data_buffer->ac3sync>16)
	  data_buffer->ac3sync=16;
	if(data_buffer->ac3sync<-1)
	  data_buffer->ac3sync=-1;
	break;
      case MENU_FILE:
	data_buffer->changeFile=i;
	usleep(300000);
	break;
      }
      display_line(dev, pos, 1, data_buffer);
    }
  }
  data_buffer->subtitleStat=SUBTITLE_DISPLAY;
}

void go_to(int dev, struct data_buf *data_buffer, int direction){

  int i=0;
  char tbuff[16];

  data_buffer->subtitleStat=SUBTITLE_HIDE;
  OSDClose(dev);
  OSDOpen(dev, 100, 40, 420, 80, 2, 0);
  OSDSetColor(dev, 0, 0, 0, 0, 255);
  OSDSetColor(dev, 1, 0, 255, 255, 255);
  OSDSetColor(dev, 2, 255, 0, 255, 255);
  if(direction==1)
    OSDText(dev, 20, 7, FONT_SIZE, 1, " FORWARD");
  else
    OSDText(dev, 20, 7, FONT_SIZE, 1, "BACKWARD");
  OSDText(dev, 170, 7, FONT_SIZE, 2, "  0");
  OSDText(dev, 240, 7, FONT_SIZE, 1, "MB");
  while(get_ir()){
    if(data_buffer->ir.state==NUM_KEY){
      i=10*i+data_buffer->ir.number;
      if(i>8*1024){
	i=8*1024;
      }
      sprintf(tbuff, "%4d", i);
      OSDText(dev, 170, 7, FONT_SIZE, 2, tbuff);
      if(i==8*1024)
	break;
    }
    else if(!strcmp(data_buffer->ir.command, ENTER))
      break;
    else if(!strcmp(data_buffer->ir.command, NORMAL)){
      i=0;
      break;
    }
  }
  direction*=i;
  data_buffer->go_to=direction;
  sleep(1);
  OSDClose(dev);
  data_buffer->subtitleStat=SUBTITLE_DISPLAY;
}
