/*
 * subpic_decode.c - converts DVD subtitles to an XPM image
 *
 * Mostly based on hard work by:
 *
 * Copyright (C) 2000   Samuel Hocevar <sam@via.ecp.fr>
 *                       and Michel Lespinasse <walken@via.ecp.fr>
 *
 * Lots of rearranging by:
 *   Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 *   Matjaz Thaler <matjaz.thaler@rd.iskraemeco.si>
 *     added support for subtitles on DVB-s PCI card
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *                                                     
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>	/* semaphore */
#include <sys/ipc.h>	/* semaphore */
#include <sys/sem.h>	/* semaphore */

#include <OSD.h>
#include "dvdplayer.h"

typedef unsigned int uint_32;
typedef unsigned char uint_8;

int data_read(unsigned char*, int);
int packet_size;
int id; /* 0: up  1: down */
int aligned;
int offset[2];
uint_8 palette[4];
uint_32 palette_to_rgb[16] = 
{
	0x00505050,
	0x00ff0000,
	0xffffffff,
	0x00000000,
	0x00005555,
	0x0000ff00,
	0x00cc2255,
	0x00cc0055,
	0x00404040,
	0x00202020,
	0x00b0b0b0,
	0x00d0d0d0,
	0x60606000,
	0x70707000,
	0x80808000,
	0x60606000,
};


unsigned char trans[4];
unsigned char buffer[10 * 1024];

int width = 0;
int height = 0;
int data_size;
unsigned char bmp[720*574*2];
char lineStat[574];

#define BUFFSIZE 16*2048

unsigned int get_byte (void)
{
	unsigned char byte;

	byte = buffer[offset[id]];
	offset[id]++;

	return byte;
}

int get_nibble (void)
{
	static int next;

	if (aligned)
	{
		next = get_byte ();
		aligned = 0;
		return next >> 4;
	}
	else
	{
		aligned = 1;
		return next & 0xf;
	}
}

uint_32 x;
uint_32 x_org;
uint_32 y;
uint_32 y_org;
uint_32 display_time;


uint_32 *image;

void initialize(void)
{
	image = malloc(720 * 576 * sizeof(uint_32)*4);

	//display_init();

#if 0

	/* FIXME : the palette isn't accurate yet,
	 *         and we don't handle transparency
	 */
	printf("/* XPM */\n");
	printf("static char *foo[] = {\n");
	printf("/* columns rows colors chars-per-pixel */\n");
	printf("\"%i %i 16 1\",\n", width, height);
	printf("\"0 c grey60\",\n");
	printf("\"1 c red\",\n");
	printf("\"2 c white\",\n");
	printf("\"3 c black\",\n");
	printf("\"4 c yellow\",\n");
	printf("\"5 c green\",\n");
	printf("\"6 c magenta\",\n");
	printf("\"7 c cyan\",\n");
	printf("\"8 c grey10\",\n");
	printf("\"9 c grey30\",\n");
	printf("\"a c grey40\",\n");
	printf("\"b c grey50\",\n");
	printf("\"c c grey60\",\n");
	printf("\"d c grey70\",\n");
	printf("\"e c grey80\",\n");
	printf("\"f c grey60");
#endif
}	

void draw_pixels(length, colorid)
{
  if(colorid != 0)
    lineStat[y]=1;
  while (length--){
    bmp[x++ + y * 720]=colorid;
  }
  //image[x++ + y * 720] = palette_to_rgb[palette[colorid]];
}

void next_line(void)
{
	draw_pixels(width - (x - x_org), 0);
	x = x_org;
	y++;
	id = 1 - id;
}

void terminate(void)
{
	while (y < (height + y_org))
		next_line();
}

void parse_subpic_header()
{
	unsigned char dummy;
	int i;

	//parse off the padding bytes. this should be done in extract_subpic
sync:	do
	  if(!data_read(&buffer[0], 1))
            sleep(1);
	while(buffer[0] == 0xff);
	
	data_read(&buffer[1], 1);
	packet_size = (buffer[0] << 8) + buffer[1];
	if(packet_size==0 || packet_size>16384){
	  //fprintf(stderr,"(subpic) packet_size = %d\n",packet_size);
	  goto sync;
	}

	i=0;
	while(i<packet_size+4)
	  i+=data_read(buffer+2+i, packet_size+4-i);
	data_size = (((unsigned int)buffer[2]) << 8) + buffer[3];
	if(packet_size < data_size){
	  //fprintf(stderr,"(subpic) packet_size = %d data_size = %d\n",packet_size,data_size);
	  goto sync;
	}

	  
#ifdef DEBUG
	fprintf(stderr,"(subpic) packet_size = %d data_size = %d\n",packet_size,data_size);
#endif DEBUG

	/* parse the header */

	i = data_size + 4;
	while (i < packet_size)
	{
		dummy = buffer [i];
		switch (dummy)
		{
			case 0x01: /* null packet ? */
			i++;
			break;

			case 0x02: /* 02 ff (ff) is the end of the packet */
			i = packet_size;
			break;

			case 0x03: /* palette */
			palette[0] = buffer [i+1] >> 4;
			palette[1] = buffer [i+1] & 0xf;
			palette[2] = buffer [i+2] >> 4;
			palette[3] = buffer [i+2] & 0xf;
#ifdef DEBUG
			fprintf(stderr,"(subpic) palette [%d %d %d %d]\n",palette[0],palette[1],
					palette[2], palette[3]);
#endif DEBUG
			i += 3;
			break;

			case 0x04: /* transparency palette */
			trans[0] = buffer [i+1] >> 4;
			trans[1] = buffer [i+1] & 0xf;
			trans[2] = buffer [i+2] >> 4;
			trans[3] = buffer [i+2] & 0xf;
#ifdef DEBUG
			fprintf(stderr,"(subpic) trans [%d %d %d %d]\n",trans[0],trans[1],
					trans[2],trans[3]);
#endif DEBUG
			i += 3;
			break;
			
			case 0x05: /* image coordinates */
			x_org = (((unsigned int)buffer[i+1]) << 4) + (buffer[i+2] >> 4);
			y_org = (((unsigned int)buffer[i+4]) << 4) + (buffer[i+5] >> 4);
			x = x_org;
			y = y_org;

			width = 1 + ( ((buffer[i+2] & 0x0f) << 8) + buffer[i+3] ) - x_org;
			height = 1 + ( ((buffer[i+5] & 0x0f) << 8) + buffer[i+6] ) - y_org;
				
#ifdef DEBUG
			fprintf(stderr,"(subpic) x_org = %d y_org = %d width = %d height = %d\n", x_org, y_org, width, height);
#endif DEBUG
			i += 7;
			break;

			case 0x06: /* image 1 / image 2 offsets */
			offset[0] = (((unsigned int)buffer[i+1]) << 8) + buffer[i+2];
			offset[1] = (((unsigned int)buffer[i+3]) << 8) + buffer[i+4];
#ifdef DEBUG
			fprintf(stderr,"(subpic) field_1_offset = %d field_2_offset = %d\n",offset[0],
					offset[1]);
#endif DEBUG
			i += 5;
			break;

			case 0xff: /* "ff xx yy zz uu" with 'zz uu' == start of control packet
			* I don't know yet what xx and yy stand for
			*/
			if ((buffer[i+3] != buffer[data_size+2]) || 
					(buffer[i+4] != buffer[data_size+3]) )
			{
#ifdef DEBUG
				fprintf(stderr, "invalid control header (%.2x%.2x != %.2x%.2x) !\n",
				buffer[i+3], buffer[i+4], buffer[data_size+2], buffer[data_size+3] );
#endif DEBUG
				//exit(1);
			}
			display_time = (buffer[i+1] << 8) + buffer[i+2];
#ifdef DEBUG
			fprintf(stderr,"(subpic) display_time = %d = %f s\n", display_time, display_time / 30.0); 
#endif DEBUG
			i += 5;
			break;

			default:
#ifdef DEBUG
			fprintf(stderr, "invalid sequence in control header (%.2x) !\n", dummy);
#endif DEBUG
			i++;
			goto sync;
			//exit(1);
		}
	}

	id = 0;
	aligned = 1;
}
void
parse_subpic_data(void)
{
	while ((offset[1] < data_size + 2))
	{
		int code;
		code = get_nibble ();
		if (code >= 0x4)	/* 4 .. f */
		{
found_code:
			draw_pixels(code >> 2, code & 3);
			if (x >= width + x_org)
			{
				if (!aligned)
					get_nibble ();
					next_line ();
			}
			continue;
		}

		code = (code << 4) + get_nibble();
		if (code >= 0x10)	/* 1x .. 3x */
			goto found_code;

		code = (code << 4) + get_nibble();
		if (code >= 0x40)	/* 04x .. 0fx */
		{
			//if (aligned) fprintf (stderr, "0%x at odd offset\n", code); /* 0 xx */
			goto found_code;
		}

		code = (code << 4) + get_nibble();
		if (code >= 0x100)	/* 01xx .. 03xx */
		{
			//if (!aligned) fprintf (stderr, "0%x at odd offset\n", code); /* 0 xx x */
			goto found_code;
		}

		/* 00xx - should only happen for 00 00 */
		if (!aligned)
			code = (code << 4) + get_nibble(); /* 0 0x xx */

		if (code)
		{
			fprintf (stderr, "got unknown code 00%x (offset %x side %x, x=%d, y=%d)\n", code, offset[id], id, x, y);
			next_line ();
			continue;
		}

		/* aligned 00 00 */
		next_line ();
	} 
}

void subtitles_decode(int videodev){

  int dev;
  unsigned char bmp1[720*574];
  int firstLine, lastLine, firstColumn, lastColumn;
  int i, j, k, w;

  dev = videodev;

  for(i=0; i<574*720; i++)
    bmp[i]=0;
  
  while(1){
      //            0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
      char r[16]={  30, 64,255, 32,154,  0,154,  0,200, 16,224, 96,154,180,204,184};
      char g[16]={  30, 64,255, 32,154,255,  0,154,200, 16,224, 96,154,180,204,184};
      char b[16]={  30, 64,255, 32,  0,  0,154,154,200, 16,224, 96,154,180,204,184};
      char c[16]={255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,164};

      for(i=0; i<574; i++)
	lineStat[i]=0;

      parse_subpic_header();
      parse_subpic_data();
      terminate();
       
      if(height>200 || y_org<420){
	firstLine=574; lastLine=0; firstColumn=720; lastColumn=0;
	for(i=0; i<573; i++){
	  if(lineStat[i]==1){
	    if(firstLine>i)
	      firstLine=i;
	    if(lastLine<i)
	      lastLine=i;
	    for(j=0; j<720; j++){
	      if(bmp[i*720+j]!=0){
		if(firstColumn>j)
		  firstColumn=j;
		if(lastColumn<j)
		  lastColumn=j;
	      }
	    }
	  }
	}
	x_org=firstColumn-15;
	width=lastColumn-firstColumn+30;
	y_org=firstLine-7;
	height=lastLine-firstLine+14;
#ifdef DEBUG
	fprintf(stderr, "firstLine=%d lastLine=%d firstColumn=%d lastColumn=%d\n", 
		firstLine, lastLine, firstColumn, lastColumn);
#endif DEBUG
      }
	
      if(data_buffer->subtitleStat==SUBTITLE_DISPLAY)
	OSDOpen(dev, x_org, y_org, x_org+width, y_org+height, 4, 0);
      if(data_buffer->subtitleStat!=SUBTITLE_HIDE){
	OSDSetColor(dev, 0, r[palette[0]], g[palette[0]], b[palette[0]], c[trans[0]]);
	OSDSetColor(dev, 1, r[palette[1]], g[palette[1]], b[palette[1]], c[trans[1]]);
	OSDSetColor(dev, 2, r[palette[2]], g[palette[2]], b[palette[2]], c[trans[2]]);
	OSDSetColor(dev, 3, r[palette[3]], g[palette[3]], b[palette[3]], c[trans[3]]);
      }
      j=0; firstLine=574; lastLine=0; firstColumn=720; lastColumn=0;
      while(j<(height-1)){
	if(lineStat[j+y_org]==1){
	  if(j<firstLine) 
	    firstLine=j;
	  if(j>lastLine) 
	    lastLine=j;
	  for(i=0; i<width; i++){
	    if(bmp[(j+y_org)*720+i+x_org]!=0){
	      if(firstColumn>i)
		firstColumn=i;
	      if(lastColumn<i)
		lastColumn=i;
	    }
	  }
	}
	if(firstLine<lastLine && lastLine!=j){
	  w=lastColumn-firstColumn;
	  for(k=0; k<=(lastLine-firstLine); k++)
	    for(i=0; i<=w; i++){
	      bmp1[i+k*w]=bmp[x_org+i+firstColumn+(k+firstLine+y_org)*720];
	      bmp[x_org+i+firstColumn+(k+firstLine+y_org)*720]=0;
	    }
	  if(data_buffer->subtitleStat==SUBTITLE_DISPLAY){
	    OSDSetBlock(dev, firstColumn, firstLine, w, lastLine, w, bmp1);
	  }
	  else if(data_buffer->subtitleStat==SUBTITLE_OVERLAY){
	    i=x_org+firstColumn-100;
	    if(i<0)
	      i=0;
	    OSDSetBlock(dev, i, firstLine+250, w, lastLine+250, w, bmp1);
	  }
	  usleep(20000);
	  firstLine=574; lastLine=0; firstColumn=720; lastColumn=0;
	}
	j++;
      }
      if(data_buffer->subtitleStat!=SUBTITLE_HIDE)
	usleep(display_time*12000);
      if(data_buffer->subtitleStat==SUBTITLE_DISPLAY)
	OSDClose(dev);
      else if(data_buffer->subtitleStat==SUBTITLE_OVERLAY)
	OSDFillBlock(dev, 0, 240, 520, 400, 0);
  }
  close(dev);
}

int data_read(unsigned char *data, int datasize)
{
  int datasize2;

  datasize2=0;
  
  while(datasize){
    if(data_buffer->output_pointer != data_buffer->input_pointer){ // data in the buffer
      data[datasize2++]=data_buffer->data[data_buffer->output_pointer];
      datasize--;
      data_buffer->output_pointer=
        (data_buffer->output_pointer+1)%SUBTITLE_BUFFER_SIZE;
    }
    else{                        // buffer empty
      if(sem_ctl(semId, 0, 2))   
	exit(-1);
      if(sem_op(semId, 0, 0))    // wait for data
	exit(-1);
    }
  }
  if(sem_ctl(semId, 0, 1))       // wake up main if sleeps
    exit(-1);
  if(data_buffer->exit_child) 
    exit(-1);
  return(datasize2);
}




